/***************************************************

Copyright (c) 1999  kashiwano masahiro

Abstract:

    UUSBD.SYS Pnp functions
    
****************************************************/

#include "wdm.h"
#include "stdarg.h"
#include "stdio.h"

#include "usbdi.h"
#include "usbdlib.h"
#include "uusbd.h"

ULONG SumStagedPendingIrpCount(PDEVICE_EXTENSION deviceExtension)
{
	ULONG i,j;
	ULONG sum;
	sum = 0;
	for(i=0;i<deviceExtension->UsbNumInterface;i++){
		for(j=0;j<deviceExtension->UsbInterfaceInfo[i].NumberOfPipes;j++){
			sum += deviceExtension->UsbInterfaceInfo[i].Pipe[j].StagedPendingIrpCount; 
		}
	}
	return sum;
}


NTSTATUS
Uusbd_ProcessPnPIrp(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP           Irp
    )
/***

Routine Description:

    Plug and Play IRP̏[`.

Arguments:
	DeviceObject
		foCXFDOւ̃|C^[

    Irp
		I/O Request Packetւ̃|C^[

Return Value:

    NT status code

***/
{

    PIO_STACK_LOCATION irpStack;
    PDEVICE_EXTENSION deviceExtension;
    NTSTATUS ntStatus = STATUS_SUCCESS;
    NTSTATUS waitStatus;
    PDEVICE_OBJECT stackDeviceObject;
    KEVENT startDeviceEvent;


    //
    // ݂IO_STACK_LOCATION𓾂
    //  function codesp[^[
    //
    irpStack = IoGetCurrentIrpStackLocation (Irp);

    //
    // Get a pointer to the device extension
    //

    deviceExtension = DeviceObject->DeviceExtension;
    stackDeviceObject = deviceExtension->TopOfStackDeviceObject;

    UUSBD_KdPrint( DBGLVL_MEDIUM, ( "enter Uusbd_ProcessPnPIrp() IRP_MJ_PNP, minor %s\n",
        UUSBD_StringForPnpMnFunc( irpStack->MinorFunction ) ));

	// countCNg
    Uusbd_IncrementIoCount(DeviceObject);

    UUSBD_ASSERT( IRP_MJ_PNP == irpStack->MajorFunction );

    switch (irpStack->MinorFunction) {
    case IRP_MN_START_DEVICE:

		// PnP}l[W̓\[X蓖ĂIRP_MN_START_DEVICE
		// 𑗂B
		// ܂wɂIRP𑗂

		// ŵm点CxgĂ
        KeInitializeEvent(&startDeviceEvent, NotificationEvent, FALSE);
        IoCopyCurrentIrpStackLocationToNext(Irp);

        // wIRP̏Uusbd_IrpCompletionRoutineĂ΂
		// 悤ɂ.̊֐̒ startDeviceEventVOiԂɂ
        IoSetCompletionRoutine(Irp,
                               Uusbd_IrpCompletionRoutine,	//[`
                               &startDeviceEvent,			//[` Context
                               TRUE,    // invoke on success
                               TRUE,    // invoke on error
                               TRUE);   // invoke on cancellation


        // wIRP
        ntStatus = IoCallDriver(stackDeviceObject,
                                Irp);

        // if PDO is not done yet, wait for the event to be set in our completion routine
        if (ntStatus == STATUS_PENDING) {
	        // w̏ĂȂꍇ܂ő҂B
			// startDeviceEventVOiԂɂȂ
            waitStatus = KeWaitForSingleObject(
                &startDeviceEvent,
                Suspended,
                KernelMode,
                FALSE,
                NULL);
			ntStatus = Irp->IoStatus.Status;
        }
		if (NT_SUCCESS(ntStatus)) {

		//̎_ŃfoCXgp\ɂȂ
		//Ȃǂs
	        ntStatus = Uusbd_StartDevice(DeviceObject);
	        Irp->IoStatus.Status = ntStatus;
		}
		IoCompleteRequest (Irp,
						   IO_NO_INCREMENT
						   );

		Uusbd_DecrementIoCount(DeviceObject);
		return ntStatus;  // end, case IRP_MN_START_DEVICE

    case IRP_MN_QUERY_STOP_DEVICE:

		// STOP_DEVICE̗v
		// The IRP_MN_QUERY_STOP_DEVICE/IRP_MN_STOP_DEVICE sequence only occurs
		// during "polite" shutdowns, such as the user explicitily requesting the
		// service be stopped in, or requesting unplug from the Pnp tray icon.
		// This sequence is NOT received during "impolite" shutdowns,
		// such as someone suddenly yanking the USB cord or otherwise 
		// unexpectedly disabling/resetting the device.

        // If a driver sets STATUS_SUCCESS for this IRP,
        // the driver must not start any operations on the device that
        // would prevent that driver from successfully completing an IRP_MN_STOP_DEVICE
        // for the device.
        // For mass storage devices such as disk drives, while the device is in the
        // stop-pending state,the driver holds IRPs that require access to the device,
        // but for most USB devices, there is no 'persistent storage', so we will just
        // refuse any more IO until restarted or the stop is cancelled

        // If a driver in the device stack determines that the device cannot be
        // stopped for resource reconfiguration, the driver is not required to pass
        // the IRP down the device stack. If a query-stop IRP fails,
        // the PnP Manager sends an IRP_MN_CANCEL_STOP_DEVICE to the device stack,
        // notifying the drivers for the device that the query has been cancelled
        // and that the device will not be stopped.


        // It is possible to receive this irp when the device has not been started
        //  ( as on a boot device )
        if (!deviceExtension->DeviceStarted) { // if get when never started, just pass on
            UUSBD_KdPrint( DBGLVL_MEDIUM,("Uusbd_ProcessPnPIrp() IRP_MN_QUERY_STOP_DEVICE when device not started\n"));
            IoSkipCurrentIrpStackLocation (Irp);
            ntStatus = IoCallDriver (deviceExtension->TopOfStackDeviceObject, Irp);
            Uusbd_DecrementIoCount(DeviceObject);

            return ntStatus;
        }


        // read/write IRP҂ĂꍇAۂ
        if( SumStagedPendingIrpCount(deviceExtension) ) {
            ntStatus = STATUS_UNSUCCESSFUL;
        }
        else { 
			// ۂȂ
			// ȌAVIOCTL  read/writeȂ悤ɃtO𗧂Ă
             deviceExtension->StopDeviceRequested = TRUE;
			 Irp->IoStatus.Status = STATUS_SUCCESS;
        }

        break; // end, case IRP_MN_QUERY_STOP_DEVICE

    case IRP_MN_CANCEL_STOP_DEVICE:

		// PnP}l[W[͒~łȂfoCXA
		// IRP𑗂~vLZ邱Ƃm点
		// IRPIRP_MN_QUERY_STOP_DEVICEfoCXɑ΂Ă̂ݑ


        // It is possible to receive this irp when the device has not been started
        if (!deviceExtension->DeviceStarted) { // ܂X^[gĂȂ́AwɃpX̂
            UUSBD_KdPrint( DBGLVL_MEDIUM,("Uusbd_ProcessPnPIrp() IRP_MN_CANCEL_STOP_DEVICE when device not started\n"));
            IoSkipCurrentIrpStackLocation (Irp);
            ntStatus = IoCallDriver (deviceExtension->TopOfStackDeviceObject, Irp);
            Uusbd_DecrementIoCount(DeviceObject);
            return ntStatus;
        }

		// ĂIOCTL  read/write\ɂ
        deviceExtension->StopDeviceRequested = FALSE;
		Irp->IoStatus.Status = STATUS_SUCCESS;
        break; // end, case IRP_MN_CANCEL_STOP_DEVICE

    case IRP_MN_STOP_DEVICE:

		// PnP Manager̓foCX~鎞ɂIRP𑗂
		// ~ƃn[hEGA\[X̍čO[V
		// \ɂȂ
		// IRP_MN_QUERY_STOP_DEVICEƂ݂̂IRP𑗂

        ntStatus = Uusbd_StopDevice(DeviceObject, TRUE);

		Irp->IoStatus.Status = ntStatus;
        break; // end, case IRP_MN_STOP_DEVICE



    case IRP_MN_QUERY_REMOVE_DEVICE:

		//IRP͈SɁiVXeɈejfoCXO
		//Ƃł邩ǂ₢킹邽߂Ɏg
		//hCo[foCXOĂSƔf
		//܂IĂȂIONGXgAȍ~read/write
		//NGXg߂STATUS_SUCCESSԂB
		//̏ꍇAIRP_MN_REMOVE_DEVICȄmɐ悤
		//ȌVĂ͂Ȃ

		//OȂƔfꍇÃfoCXX^bN
		//IRP_MN_QUERY_REMOVE_DEVICE𑗂Ă͂ȂB
		//IRPsꍇAPnP}l[W[́A
		//IRP_MN_CANCEL_REMOVE_DEVICEfoCXX^bNɑ
		//REMOVE_DEVICE̗vLZ

        // It is possible to receive this irp when the device has not been started
        if (!deviceExtension->DeviceStarted) { // if get when never started, just pass on
            UUSBD_KdPrint( DBGLVL_MEDIUM,("Uusbd_ProcessPnPIrp() IRP_MN_QUERY_STOP_DEVICE when device not started\n"));
            IoSkipCurrentIrpStackLocation (Irp);
            ntStatus = IoCallDriver (deviceExtension->TopOfStackDeviceObject, Irp);
            Uusbd_DecrementIoCount(DeviceObject);

            return ntStatus;
        }

		// ȌAVIOCTL  read/writeIRPw̃hCo[ɓnȂ
		// 悤ɂ悤ɃtO𗧂Ă
       deviceExtension->RemoveDeviceRequested = TRUE;

		// ̃[`̏߂ŃCNgĂ̂ňꎞIɃfNg
		Uusbd_DecrementIoCount(DeviceObject);

        // yfBOIONGXg΂ꂪI܂ł܂
		// deviceExtension->PendingIoCount1ɂȂANoPendingIoEvent
		// VOiԂɂȂ
        waitStatus = KeWaitForSingleObject(
                    &deviceExtension->NoPendingIoEvent,
                    Suspended,
                    KernelMode,
                    FALSE,
                    NULL);

		// ꎞIȃfNg߂		
		Uusbd_IncrementIoCount(DeviceObject);

		Irp->IoStatus.Status = STATUS_SUCCESS;
        break; // end, case IRP_MN_QUERY_REMOVE_DEVICE

    case IRP_MN_CANCEL_REMOVE_DEVICE:

		// PnP}l[W[uvƂIRPg
		// IRP_MN_QUERY_REMOVE_DEVICEƂ݂̂IRP͗

        if (!deviceExtension->DeviceStarted) { // if get when never started, just pass on
            UUSBD_KdPrint( DBGLVL_MEDIUM,("Uusbd_ProcessPnPIrp() IRP_MN_CANCEL_REMOVE_DEVICE when device not started\n"));
            IoSkipCurrentIrpStackLocation (Irp);
            ntStatus = IoCallDriver (deviceExtension->TopOfStackDeviceObject, Irp);
            Uusbd_DecrementIoCount(DeviceObject);
            return ntStatus;
        }

		// ĂсAVIOCTL  IO Irp 󂯂悤ɂ
        deviceExtension->RemoveDeviceRequested = FALSE;

		Irp->IoStatus.Status = STATUS_SUCCESS;
        break; // end, case IRP_MN_CANCEL_REMOVE_DEVICE

	case IRP_MN_SURPRISE_REMOVAL:
        UUSBD_KdPrint( DBGLVL_MEDIUM,("Uusbd_ProcessPnPIrp() IRP_MN_SURPRISE_REMOVAL\n"));
		
		// ˑRfoCXOꂽꍇ 
		// ̏ꍇPnP}l[W[IRP_MN_QUERY_REMOVE_DEVICE
		// 邱ƂȂۂ̃foCXɎOĂ

        // match the inc at the begining of the dispatch routine
        Uusbd_DecrementIoCount(DeviceObject);

		//
        // ȍ~AVIOCTLread/write irpw̃hCo[ɓnȂ
		// 悤ɃtO𗧂Ă
        deviceExtension->DeviceRemoved = TRUE;

		if(deviceExtension->DeviceStarted){
			Uusbd_StopDevice(DeviceObject, FALSE);
		}

		Irp->IoStatus.Status = STATUS_SUCCESS;
		// We don't explicitly wait for the below driver to complete, but just make
		// the call and go on, finishing cleanup
		// w̃hCo[IRP𑗂
        IoCopyCurrentIrpStackLocationToNext(Irp);

        ntStatus = IoCallDriver(stackDeviceObject,
                                Irp);

		return ntStatus;

    case IRP_MN_REMOVE_DEVICE:

		// foCXO PnP Manager͂IRP𔭍s
		// \ꂽȌꍇAOIRP_MN_QUERY_REMOVE_DEVICE
		// foCX̓[u҂ԂɂȂĂB
		// \ȂꍇiR[h𔲂ꍇjȂ̑OG
		// Ȃۂ̃foCX̂̂OĂB
		// ̏ꍇ́AIRP_MN_QUERY_REMOVE_DEVICE͑ȂB

        // match the inc at the begining of the dispatch routine
        Uusbd_DecrementIoCount(DeviceObject);

		//
        // ȍ~AVIOCTLread/write irpw̃hCo[ɓnȂ
		// 悤ɃtO𗧂Ă
        deviceExtension->DeviceRemoved = TRUE;

		if(deviceExtension->DeviceStarted){
			Uusbd_StopDevice(DeviceObject, TRUE);
		}
		// We don't explicitly wait for the below driver to complete, but just make
		// the call and go on, finishing cleanup
		// w̃hCo[IRP𑗂
        IoCopyCurrentIrpStackLocationToNext(Irp);

        ntStatus = IoCallDriver(stackDeviceObject,
                                Irp);
		// PendingIoCountfNgB
		// ŁAyfBOIRPȂPendingIoCount0ɂȂ
		// 0ɂȂɁARemoveEventVOiԂɂȂ
        Uusbd_DecrementIoCount(DeviceObject);


        // wait for any io request pending in our driver to
        // complete for finishing the remove

        KeWaitForSingleObject(
                    &deviceExtension->RemoveEvent,
                    Suspended,
                    KernelMode,
                    FALSE,
                    NULL);

        //
        // Delete the link and FDO we created
        //
        Uusbd_RemoveDevice(DeviceObject);

        return ntStatus; // end, case IRP_MN_REMOVE_DEVICE


    default:
        UUSBD_KdPrint( DBGLVL_MAXIMUM,("Uusbd_ProcessPnPIrp() Minor PnP IOCTL not handled\n"));
    } /* case MinorFunction  */


    if (!NT_SUCCESS(ntStatus)) {

		// G[ꍇAwIRP𑗂炸I
        Irp->IoStatus.Status = ntStatus;
		IoCompleteRequest (Irp,
						   IO_NO_INCREMENT
						   );

		Uusbd_DecrementIoCount(DeviceObject);

		UUSBD_KdPrint( DBGLVL_MAXIMUM,("Uusbd_ProcessPnPIrp() Exit Uusbd_ProcessPnPIrp FAILURE %x\n", ntStatus));
		return ntStatus;
    }

    //
    // SĂ PNP_POWER messages͉w̃foCXIuWFNgiTopOfStackDeviceObjectj
	// ɑ
	// IRP_MN_START_DEVICȄꍇ͉wɑĂAg̃foCXɊւ鏈
	// ŝŁAȉ̃[`͎gpȂB
	// Ȍꍇ͎g̃foCX̏Ăȉ̃[`g
	// wIRP
    IoCopyCurrentIrpStackLocationToNext(Irp);

    //
    // All PNP_POWER messages get passed to the TopOfStackDeviceObject
    // we were given in PnPAddDevice
    //

    UUSBD_KdPrint( DBGLVL_MAXIMUM,("Uusbd_ProcessPnPIrp() Passing PnP Irp down, status = %x\n", ntStatus));

    ntStatus = IoCallDriver(stackDeviceObject,
                            Irp);

    Uusbd_DecrementIoCount(DeviceObject);

    UUSBD_KdPrint( DBGLVL_MAXIMUM,("Uusbd_ProcessPnPIrp() Exit Uusbd_ProcessPnPIrp %x\n", ntStatus));

    return ntStatus;
}


NTSTATUS
Uusbd_PnPAddDevice(
    IN PDRIVER_OBJECT DriverObject,
    IN PDEVICE_OBJECT PhysicalDeviceObject
    )
/***

Routine Description:

	Functional Device Object (FDO)쐬鎞ɂт.

Arguments:

    DriverObject
		UUSBD̃foCXIuWFNg

    PhysicalDeviceObject
		̃foCXX^bNPDOioXhCo[ɂčj

Return Value:

    STATUS_SUCCESS if successful,
    STATUS_UNSUCCESSFUL otherwise

***/
{
    NTSTATUS                ntStatus = STATUS_SUCCESS;
    PDEVICE_OBJECT          deviceObject = NULL;
    PDEVICE_EXTENSION       deviceExtension;
    USBD_VERSION_INFORMATION versionInformation;
    ULONG i;


    
    UUSBD_KdPrint( DBGLVL_DEFAULT,("enter Uusbd_PnPAddDevice()\n"));

    //
    // create our funtional device object (FDO)
    //

    // funtional device object (FDO)
    ntStatus =
        Uusbd_CreateDeviceObject(DriverObject, PhysicalDeviceObject, &deviceObject);

    if (NT_SUCCESS(ntStatus)) {
        deviceExtension = deviceObject->DeviceExtension;

//        deviceObject->Flags &= ~DO_DEVICE_INITIALIZING;

        // read/writeA[U[[h[
		// ڃf[^[ direct io T|[g 
        deviceObject->Flags |= DO_DIRECT_IO;


		// TXyhAW[IRP_MN_STOP_DEVICE,IRP_MN_START_DEVICE
		// IRPNȂ悤ɂ
		// IRP_MN_START_DEVICEGetDescriptors()ŝh߁B
        deviceObject->Flags |= DO_POWER_PAGABLE;
        // device extension̏
        //
        // Physical device ObjectoĂ
        deviceExtension->PhysicalDeviceObject=PhysicalDeviceObject;

        //
        // PDOɃA^b`B
        // A^b`ɈԏɂfoCX
		// TopOfStackDeviceObjectɊoĂ
		// ꂪ̉wɂȂB

        deviceExtension->TopOfStackDeviceObject =
            IoAttachDeviceToDeviceStack(deviceObject, PhysicalDeviceObject);

		// device extensionDEVICE_CAPABILITIESɃfoCX̕I
		// Rs[
        // We are most interested in learning which system power states
        // are to be mapped to which device power states for handling
        // IRP_MJ_SET_POWER Irps.
        Uusbd_QueryCapabilities(deviceExtension->TopOfStackDeviceObject,
                                 &deviceExtension->DeviceCapabilities);


		// We want to determine what level to auto-powerdown to; This is the lowest
		//  sleeping level that is LESS than D3; 
		// If all are set to D3, auto powerdown/powerup will be disabled.

        deviceExtension->PowerDownLevel = PowerDeviceUnspecified; // init to disabled
        for (i=PowerSystemSleeping1; i<= PowerSystemSleeping3; i++) {

			if ( deviceExtension->DeviceCapabilities.DeviceState[i] < PowerDeviceD3 )
				deviceExtension->PowerDownLevel = deviceExtension->DeviceCapabilities.DeviceState[i];
        }

#if DBG

        // May want override auto power-down level from registry;
        // ( CurrentControlSet\Services\BulkUsb\Parameters )
		// Setting to 0 or 1 in registry disables auto power-down
        Uusbd_GetRegistryDword( UUSBD_REGISTRY_PARAMETERS_PATH,
                                         L"PowerDownLevel",
                                         &(deviceExtension->PowerDownLevel) );



        //
        // display the device  caps
        //

        UUSBD_KdPrint( DBGLVL_MEDIUM,(" >>>>>> DeviceCaps\n"));
        UUSBD_KdPrint( DBGLVL_MEDIUM,(" SystemWake = %s\n",
            UUSBD_StringForSysState( deviceExtension->DeviceCapabilities.SystemWake ) ));
        UUSBD_KdPrint( DBGLVL_MEDIUM,(" DeviceWake = %s\n",
            UUSBD_StringForDevState( deviceExtension->DeviceCapabilities.DeviceWake) ));

        for (i=PowerSystemUnspecified; i< PowerSystemMaximum; i++) {

            UUSBD_KdPrint( DBGLVL_MEDIUM,(" Device State Map: sysstate %s = devstate %s\n",
                 UUSBD_StringForSysState( i ),
                 UUSBD_StringForDevState( deviceExtension->DeviceCapabilities.DeviceState[i] ) ));
        }
        UUSBD_KdPrint( DBGLVL_MEDIUM,(" <<<<<<<<DeviceCaps\n"));
#endif
        // We keep a pending IO count ( extension->PendingIoCount )  in the device extension.
        // The first increment of this count is done on adding the device.
        // Subsequently, the count is incremented for each new IRP received and
        // decremented when each IRP is completed or passed on.

        // Transition to 'one' therefore indicates no IO is pending and signals
        // deviceExtension->NoPendingIoEvent. This is needed for processing
        // IRP_MN_QUERY_REMOVE_DEVICE

        // Transition to 'zero' signals an event ( deviceExtension->RemoveEvent )
        // to enable device removal. This is used in processing for IRP_MN_REMOVE_DEVICE
        //
        Uusbd_IncrementIoCount(deviceObject);

    }

    USBD_GetUSBDIVersion(&versionInformation);



    if( NT_SUCCESS( ntStatus ) )  
    {
        NTSTATUS actStat;
        // try to power down device until IO actually requested
        actStat = Uusbd_SelfSuspendOrActivate( deviceObject, TRUE );
		deviceObject->Flags &= ~DO_DEVICE_INITIALIZING;
    }

    UUSBD_KdPrint( DBGLVL_DEFAULT,("Flag = %08x\n", deviceObject->Flags));

    UUSBD_KdPrint( DBGLVL_DEFAULT,("exit Uusbd_PnPAddDevice() (%x)\n", ntStatus));

    return ntStatus;
}


NTSTATUS
Uusbd_StartDevice(
    IN  PDEVICE_OBJECT DeviceObject
    )
/***

Routine Description:

    Uusbd_ProcessPnPIrp()̒ĂяoB
	IRP̎ނIRP_MN_START_DEVICEŁAwɂIRP̏n
	wX^[gƂ̊֐Ă΂B
	̎_ŁAURB艺wɓnƂłB
	ŁARtBO[Vs

Arguments:
    DriverObject
		UUSBD̃foCXIuWFNg

Return Value:

    NT status code

***/
{
    PDEVICE_EXTENSION deviceExtension;
    NTSTATUS ntStatus;
    PUSB_DEVICE_DESCRIPTOR deviceDescriptor = NULL;
    PURB urb;
    ULONG siz;

    UUSBD_KdPrint( DBGLVL_DEFAULT,("enter Uusbd_StartDevice\n"));

    deviceExtension = DeviceObject->DeviceExtension;

    urb = UUSBD_ExAllocatePool(NonPagedPool,
                         sizeof(struct _URB_CONTROL_DESCRIPTOR_REQUEST));

    UUSBD_KdPrintCond( DBGLVL_HIGH,!urb, ("Uusbd_StartDevice() FAILED UUSBD_ExAllocatePool() for URB\n"));

    if (urb) {

        siz = sizeof(USB_DEVICE_DESCRIPTOR);

        deviceDescriptor = UUSBD_ExAllocatePool(NonPagedPool,
                                          siz);

        UUSBD_KdPrintCond( DBGLVL_HIGH, !deviceDescriptor, ("Uusbd_StartDevice() FAILED UUSBD_ExAllocatePool() for deviceDescriptor\n"));

        if (deviceDescriptor) {

            UsbBuildGetDescriptorRequest(urb,
                                         (USHORT) sizeof (struct _URB_CONTROL_DESCRIPTOR_REQUEST),
                                         USB_DEVICE_DESCRIPTOR_TYPE,
                                         0,
                                         0,
                                         deviceDescriptor,
                                         NULL,
                                         siz,
                                         NULL);


            ntStatus = Uusbd_CallUSBD(DeviceObject, urb);

            UUSBD_KdPrintCond( DBGLVL_DEFAULT, !NT_SUCCESS(ntStatus), ("Uusbd_StartDevice() FAILED Uusbd_CallUSBD(DeviceObject, urb)\n"));

            if (NT_SUCCESS(ntStatus)) {
                UUSBD_KdPrint( DBGLVL_MEDIUM,("Device Descriptor = %x, len %x\n",
                                deviceDescriptor,
                                urb->UrbControlDescriptorRequest.TransferBufferLength));

                UUSBD_KdPrint( DBGLVL_MEDIUM,("Device Descriptor:\n"));
                UUSBD_KdPrint( DBGLVL_MEDIUM,("-------------------------\n"));
                UUSBD_KdPrint( DBGLVL_MEDIUM,("bLength %d\n", deviceDescriptor->bLength));
                UUSBD_KdPrint( DBGLVL_MEDIUM,("bDescriptorType 0x%x\n", deviceDescriptor->bDescriptorType));
                UUSBD_KdPrint( DBGLVL_MEDIUM,("bcdUSB 0x%x\n", deviceDescriptor->bcdUSB));
                UUSBD_KdPrint( DBGLVL_MEDIUM,("bDeviceClass 0x%x\n", deviceDescriptor->bDeviceClass));
                UUSBD_KdPrint( DBGLVL_MEDIUM,("bDeviceSubClass 0x%x\n", deviceDescriptor->bDeviceSubClass));
                UUSBD_KdPrint( DBGLVL_MEDIUM,("bDeviceProtocol 0x%x\n", deviceDescriptor->bDeviceProtocol));
                UUSBD_KdPrint( DBGLVL_MEDIUM,("bMaxPacketSize0 0x%x\n", deviceDescriptor->bMaxPacketSize0));
                UUSBD_KdPrint( DBGLVL_MEDIUM,("idVendor 0x%x\n", deviceDescriptor->idVendor));
                UUSBD_KdPrint( DBGLVL_MEDIUM,("idProduct 0x%x\n", deviceDescriptor->idProduct));
                UUSBD_KdPrint( DBGLVL_MEDIUM,("bcdDevice 0x%x\n", deviceDescriptor->bcdDevice));
                UUSBD_KdPrint( DBGLVL_MEDIUM,("iManufacturer 0x%x\n", deviceDescriptor->iManufacturer));
                UUSBD_KdPrint( DBGLVL_MEDIUM,("iProduct 0x%x\n", deviceDescriptor->iProduct));
                UUSBD_KdPrint( DBGLVL_MEDIUM,("iSerialNumber 0x%x\n", deviceDescriptor->iSerialNumber));
                UUSBD_KdPrint( DBGLVL_MEDIUM,("bNumConfigurations 0x%x\n", deviceDescriptor->bNumConfigurations));
            }
        } else {
			// if we got here we failed to allocate deviceDescriptor
            ntStatus = STATUS_INSUFFICIENT_RESOURCES;
        }

        if (NT_SUCCESS(ntStatus)) {
            deviceExtension->UsbDeviceDescriptor = deviceDescriptor;
        } else if (deviceDescriptor) {
            UUSBD_ExFreePool(deviceDescriptor);
        }

        UUSBD_ExFreePool(urb);

    } else {
		// if we got here we failed to allocate the urb
        ntStatus = STATUS_INSUFFICIENT_RESOURCES;
    }

    if (NT_SUCCESS(ntStatus)) {
        ntStatus = Uusbd_ConfigureDevice(DeviceObject);

        UUSBD_KdPrintCond( DBGLVL_MEDIUM,!NT_SUCCESS(ntStatus),("Uusbd_StartDevice Uusbd_ConfigureDevice() FAILURE (%x)\n", ntStatus));
    }


    if (NT_SUCCESS(ntStatus)) {
        deviceExtension->DeviceStarted = TRUE;
    }
    UUSBD_KdPrint( DBGLVL_DEFAULT, ("exit Uusbd_StartDevice (%x)\n", ntStatus));

    return ntStatus;
}

void
Uusbd_FreeMem(
    IN  PDEVICE_OBJECT DeviceObject
    )
/***

Routine Description:

	Uusbd_StartDeviceŃAP[g[

Arguments:

	DeviceObject
		Uusbd̃foCXIuWFNg

Return Value:

    

***/
{
    PDEVICE_EXTENSION deviceExtension;
	ULONG i,j;

    deviceExtension = DeviceObject->DeviceExtension;

    UUSBD_KdPrint( DBGLVL_HIGH,("Enter UUSBD_FreeMem() gExAllocCount = dec %d\n", gExAllocCount ));
    // Free device descriptor structure
    if (deviceExtension->UsbDeviceDescriptor) {
        UUSBD_ExFreePool(deviceExtension->UsbDeviceDescriptor);
		deviceExtension->UsbDeviceDescriptor = NULL;
    }

    // Free pipe info structs
    if( deviceExtension->UsbInterfaceInfo ) { 
		for(i=0;i<deviceExtension->UsbNumInterface;i++){
			PUUSBD_INTERFACEINFO interfaceinfo = &deviceExtension->UsbInterfaceInfo[i];
			if(interfaceinfo) {
				if(interfaceinfo->Pipe) {
					UUSBD_ExFreePool(interfaceinfo->Pipe);
					interfaceinfo->Pipe = NULL;
				}
			}
		}
		UUSBD_ExFreePool(deviceExtension->UsbInterfaceInfo);
		deviceExtension->UsbInterfaceInfo = NULL;
    }

	// free up the USB config discriptor
    if (deviceExtension->UsbConfigurationDescriptor) {
        UUSBD_ExFreePool(deviceExtension->UsbConfigurationDescriptor);
		deviceExtension->UsbConfigurationDescriptor = NULL;
    }
    UUSBD_KdPrint( DBGLVL_HIGH,("exit UUSBD_FreeMem() gExAllocCount = dec %d\n", gExAllocCount ));
}

NTSTATUS
Uusbd_RemoveDevice(
    IN  PDEVICE_OBJECT DeviceObject
    )
/***

Routine Description:

	foCX[uB
	AP[g[AV{bNN

Arguments:

	DeviceObject
		UUSBD̃foCXIuWFNg

Return Value:

    NT status code from free symbolic link operation

***/
{
    PDEVICE_EXTENSION deviceExtension;
    NTSTATUS ntStatus = STATUS_SUCCESS;
    UNICODE_STRING deviceLinkUnicodeString;

    UUSBD_KdPrint( DBGLVL_DEFAULT,("enter Uusbd_RemoveDevice\n"));

    deviceExtension = DeviceObject->DeviceExtension;

    RtlInitUnicodeString (&deviceLinkUnicodeString,
                          deviceExtension->DeviceLinkNameBuffer);

    // remove the GUID-based symbolic link
    ntStatus = IoSetDeviceInterfaceState(&deviceLinkUnicodeString, FALSE);
    UUSBD_ASSERT( NT_SUCCESS( ntStatus ) );


	UUSBD_KdPrint( DBGLVL_DEFAULT,("UUSBD_RemoveDevice() Detaching from %08X\n",
		deviceExtension->TopOfStackDeviceObject));

	IoDetachDevice(deviceExtension->TopOfStackDeviceObject);
	
	UUSBD_KdPrint( DBGLVL_DEFAULT,("UUSBD_RemoveDevice() Deleting %08X\n",
		DeviceObject));
	
	IoDeleteDevice (DeviceObject);

    UUSBD_KdPrint( DBGLVL_DEFAULT,("exit Uusbd_RemoveDevice() status = 0x%x\n", ntStatus ));

    return ntStatus;
}




NTSTATUS
Uusbd_UnconfigDevice(
    IN  PDEVICE_OBJECT DeviceObject
    )
/*++

Routine Description:

	USBfoCXARtBO[V̏Ԃɂ

Arguments:

	DeviceObject
		UUSBD̃foCXIuWFNg

Return Value:

    NT status code

--*/
{
    PDEVICE_EXTENSION deviceExtension;
    NTSTATUS ntStatus = STATUS_SUCCESS;
    PURB urb;
    ULONG siz;

    UUSBD_KdPrint( DBGLVL_DEFAULT,("enter Uusbd_UnconfigDevice\n"));

    deviceExtension = DeviceObject->DeviceExtension;

	// RtBO[VnhNULLɂSelect configurationĂяo
	// USBfoCX'unconfigured'stateɂȂ

    siz = sizeof(struct _URB_SELECT_CONFIGURATION);

    urb = UUSBD_ExAllocatePool(NonPagedPool,
                         siz);

    if (urb) {
        UsbBuildSelectConfigurationRequest(urb,
                                          (USHORT) siz,
                                          NULL);

        ntStatus = Uusbd_CallUSBD(DeviceObject, urb);

        UUSBD_KdPrintCond( DBGLVL_DEFAULT,!NT_SUCCESS(ntStatus),("Uusbd_UnconfigDevice() FAILURE Configuration Closed status = %x usb status = %x.\n", ntStatus, urb->UrbHeader.Status));

        UUSBD_ExFreePool(urb);
    } else {
        ntStatus = STATUS_INSUFFICIENT_RESOURCES;
    }

    UUSBD_KdPrint( DBGLVL_DEFAULT,("exit Uusbd_UnconfigDevice() (%x)\n", ntStatus));

    return ntStatus;
}

NTSTATUS
Uusbd_StopDevice(
    IN  PDEVICE_OBJECT DeviceObject,
	IN  BOOLEAN do_unconfig
    )

{
    PDEVICE_EXTENSION deviceExtension;
    deviceExtension = DeviceObject->DeviceExtension;
	// NGXgIRPLZ
	Uusbd_CancelPendingIo( DeviceObject ); 
	
	// I[vĂpCvꍇA
	// USBD  URB_FUNCTION_ABORT_PIPE𑗂ApCvN[Y
	Uusbd_AbortPipes( DeviceObject );
	
	if(do_unconfig) Uusbd_UnconfigDevice(DeviceObject);

	Uusbd_FreeMem(DeviceObject);

	deviceExtension->DeviceStarted = FALSE;
    deviceExtension->StopDeviceRequested = FALSE;

	return STATUS_SUCCESS;
}

NTSTATUS
Uusbd_IrpCompletionRoutine(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp,
    IN PVOID Context
    )
/***

Routine Description:

	// ʒm[`
	// IRPƂContextɂ킽ꂽCxgVOi
	// ɂ
	// ̃[`́APnP  Power Management̃[`Ŏgp

Arguments:

    DeviceObject
		Pointer to the device object for the class device.

    Irp
		Irp completed.

    Context
		Driver defined context, in this case a pointer to an event.

Return Value:

    The function value is the final status from the operation.

***/
{
    PKEVENT event = Context;

    // Set the input event
    KeSetEvent(event,
               1,       // Priority increment  for waiting thread.
               FALSE);  // Flag this call is not immediately followed by wait.

    // IRPIoFreeIrp()ĂяoĂȂ̂ STATUS_MORE_PROCESSING_REQUIRED
	// ԂB
    return STATUS_MORE_PROCESSING_REQUIRED;

}


